/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "auth_server.hh"
#include "socket_api.hh"
#include "buffer.hh"
#include "lcl_array.hh"
#include "sysexcept.hh"
#include <sasl/saslutil.h>
#include <iostream>
#include <cstdio>

#define USER_VALID "valid-user"
#define USER_ANON  "anonymous"

using namespace sends;
using namespace std;

//======================================  Sasl callbacks.
extern "C" {

    static int
    sasl_my_log(void* context __attribute__((unused)), 
		int prio, const char* msg) {
	if (!msg) return SASL_BADPARAM;
	printf("SASL: %s\n", msg);
	return SASL_OK;
    }

    static int
    getpath(void* context __attribute__((unused)), const char** path) {
	if (!path) return SASL_BADPARAM;
#ifdef __sun__
#warning hard-coded solaris sasl plugin path /usr/lib/sasl
	*path = "/usr/lib/sasl";
#else
#ifdef __x86_64
#warning hard-coded linux sasl plugin path /usr/lib64/sasl2
	*path = "/usr/lib64/sasl2";
#else
#warning hard-coded linux sasl plugin path /usr/lib/sasl2
	*path = "/usr/lib/sasl2";
#endif
#endif
	return SASL_OK;
    }

    static int
    getrealm(void* context __attribute__((unused)), const char** realm) {
	if (!realm) return SASL_BADPARAM;
	*realm = auth_server::refGlobal().realm();
	return SASL_OK;
    }

    static sasl_callback_t callbacks[] = {
	{SASL_CB_LOG,      (int(*)())&sasl_my_log, NULL},
	{SASL_CB_GETREALM, (int(*)())&getrealm,    NULL},
	{SASL_CB_GETPATH,  (int(*)())&getpath,     NULL},
	{SASL_CB_LIST_END, NULL,                   NULL}
    };
}


//======================================  global sasl system devic
sasl_server auth_server::global;

//======================================  global sasl constructor
sasl_server::sasl_server(void)
    : mInitialized(false)
{}

//======================================  Go away at end
sasl_server::~sasl_server(void) {
    if (mInitialized) sasl_done();
}

//======================================  Add a protocol
void 
sasl_server::addProtocol(const std::string& p) {
    if (mMechanism.empty()) {
	mMechanism = p;
    } else {
	mMechanism += " ";
	mMechanism += p;
    }
}

//======================================  Set the fully qualified domain name
void 
sasl_server::setFQDN(const std::string& d) {
    mFQDN = d;
}

//======================================  Set the Realm
void 
sasl_server::setRealm(const std::string& r) {
    mRealm = r;
}

//======================================  Set the service name
void 
sasl_server::setServer(const std::string& p) {
    mServer = p;
}

//======================================  Initialize sasl
void 
sasl_server::initialize(void) {
    if (mInitialized) return;
    thread::semlock lock(mux, thread::semlock::wrlock);
    if (mInitialized) return;

    int result=sasl_server_init(callbacks,        /* Callbacks supported */
				mServer.c_str()); /* Name of the application */

    if (result == SASL_OK) {
	mInitialized = true;
    } else {
	throw runtime_error("Unable to initialize SASL server");
    }
}

//======================================  Constructor
auth_server::auth_server(auth_type at) 
    : mAuthType(at), mContext(0)
{
}

//======================================  Destructor
auth_server::~auth_server(void) {
    close();
}

//======================================  Authenticate client on socket s
void 
auth_server::authenticate(socket_api& s, buffer& b) {
    //----------------------------------  Blow away existing data, j.i.c.
    mAuthUser.erase();

    //----------------------------------  Make sure SASL has been initialized
    if (mAuthType == kSasl) {
	auth_sasl(s,b);
    } else if (User_List.known(USER_ANON)) {
	mAuthUser = User_List[USER_ANON];
    } else {
	throw runtime_error("No anonymous user in non-authenticated server");
    }
}

//======================================  Authenticate client using sasl
void 
auth_server::auth_sasl(socket_api& s, buffer& b) {
    //----------------------------------  Make sure SASL has been initialized
    global.initialize();

    //----------------------------------  Make a context for this connection.
    open();

    int result = SASL_OK;
    try {
	const char *out;
	unsigned outlen;
	std::string mech = getMechanismList();

	// cerr << "auth_server: Mechanisms are " << out << endl;
	put(s, mech.c_str(), mech.size());
	get(s, b);

	result=sasl_server_start(mContext, /* context */
				 b.data(),
				 NULL, /* optional string from the client */
				 0,    /* and it's length */
				 &out,     /* The output of the library.
					      Might not be NULL terminated */
				 &outlen
				 );

	//------------------------------  Loop over steps.
	while (result == SASL_CONTINUE) {
	    if (outlen) put(s, out, outlen);
	    get(s, b);
	    result = sasl_server_step(mContext,
				      b.data(), /* what the client gave */
				      b.size(), /* it's length */
				      &out,     /* allocated by library*/
				      &outlen
				      );
	}


	//------------------------------  Check for successful authentication.
	if (result != SASL_OK) {
	    cerr << sasl_errdetail(mContext) << endl;
	    throw runtime_error("auth_server: authentication failure");
	}

	//------------------------------  One last turn of the crank
	if (out) {
	    put(s, out, outlen);
	    get(s, b);
	    result = sasl_server_step(mContext,
				      b.data(), /* what the client gave */
				      b.size(), /* it's length */
				      &out,     /* allocated by library*/
				      &outlen
				      );
	    if (result != SASL_OK) {
		cerr << sasl_errdetail(mContext) << endl;
		throw runtime_error("auth_server: authentication failure");
	    }
	}

	//------------------------------  Get the user name, data.
	const void* userpar = 0;
	if (sasl_getprop(mContext, SASL_USERNAME, &userpar) != SASL_OK ||
	    !userpar) {
	    throw runtime_error("No username specified");
	}
	string user(reinterpret_cast<const char*>(userpar));
	cout << "auth_server: authenticated user " << user << endl;
	if (User_List.known(user)) {
	    mAuthUser = User_List[user];
	} else if (User_List.known(USER_VALID)) {
	    cout << "auth_server: User " << user << " aliased to " 
		 << USER_VALID << "." << endl;
	    mAuthUser = User_List[USER_VALID];
	} else {
	    throw runtime_error(string("Unknown user: ")+user);
	}

    } catch (exception& e) {
	close();
	throw;
    }
}

//======================================  Authenticate client on socket s
void 
auth_server::close(void) {
    if (mContext) sasl_dispose(&mContext);
    mAuthUser.erase();
}

//======================================  Add a protocol
std::string
auth_server::getMechanismList(void) const {
    const char *out;
    unsigned outlen;
    int nMech  = 0;

    int result=sasl_listmech(mContext, /* The context for this connection */
			     NULL,     /* not supported */
			     NULL,      /* What to prepend the string with */
			     " ",     /* What to separate mechanisms with */
			     NULL,      /* What to append to the string */
			     &out,     /* The produced string. */
			     &outlen,  /* length of the string */
			     &nMech    /* # of mechanisms in string */
			     );
    if (result != SASL_OK || !nMech) {
	cerr << "SASL details: " << sasl_errdetail(mContext) << endl;
	throw runtime_error("No valid authentication mechanisms");
    }
    string mech(out, outlen);
    string mechList = refGlobal().getMechanism();
    if (!mechList.empty()) {
	string::size_type inx = 0;
	for (int i=0; i<nMech; i++) {
	    string::size_type ispc = mech.find(' ', inx);
	    if (ispc == string::npos) ispc = mech.size();
	    std::string imech = mech.substr(inx, ispc-inx);
	    while (ispc < mech.size() && mech[ispc] == ' ') ispc++;
	    
	    //--------------------------  Remove mechanism if not allowed
	    if (mechList.find(imech) == string::npos) {
		mech.erase(inx, ispc-inx);
	    }

	    //--------------------------  Bump pointer past allowed mechanism
	    else {
		inx = ispc;
	    }
	}
    }
    return mech;
}

//======================================  Authenticate client on socket s
void 
auth_server::open(void) {
    int result=sasl_server_new(global.service(), /* Registered service name */
			       global.fqdn(), /* fully qualified domain name; 
					         NULL says use gethostname() */
			       NULL, /* The user realm used for password
					lookups; NULL = default to serverFQDN
					Note: This does not affect Kerberos */
			       NULL, NULL, /* IP Address information strings */
			       NULL, /* Callbacks for this connection */
			       0, /* security flags (security layers are enabled
				   * using security properties, separately) */
			       &mContext
			       );

    if (result != SASL_OK) {
	throw runtime_error("Unable to create SASL connection context.");
    }

    sasl_security_properties_t secprops;
    memset(&secprops, 0L, sizeof(secprops));
    secprops.maxbufsize = 2048;
    secprops.min_ssf    = 0;
    secprops.max_ssf    = 256;
    secprops.security_flags = SASL_SEC_NOPLAINTEXT | SASL_SEC_NOANONYMOUS;

    result = sasl_setprop(mContext,
			  SASL_SEC_PROPS,
			  &secprops
			  );

    if (result != SASL_OK) throw runtime_error("Setting security properties");
}

//======================================  Read a data buffer
void 
auth_server::get(socket_api& s, buffer& b) {
    b.clear();
    int len = s.gets(b.data(), b.capacity());
    if (len < 0) {
	string msg = "auth_server::get() error in ";
	msg += s.where();
	throw system_error(msg, s.sys_errno());
    }
    b.push(len);
    unsigned int raw_len = 0;
    sasl_decode64(b.data(), len, b.data(), b.capacity(), &raw_len);
    b.resize(raw_len);
}

//======================================  Put a data buffer.
void 
auth_server::put(socket_api& s, const char* out, unsigned int outlen) {
    int l64 = ((outlen / 3) + 1) * 4 + 2;
    lcl_array<char> enc_str(l64);
    unsigned int enc_len = 0;
    sasl_encode64(out, outlen, enc_str, l64, &enc_len);
    enc_str[enc_len++] = '\n';
    if (s.write(enc_str, enc_len) < 0) {
	string msg = "auth_server::put() error in ";
	msg += s.where();
	throw system_error(msg, s.sys_errno());
    }
}
