/* -*- mode: c++; c-basic-offset: 4; -*- */

/**  This program merges channel lists for a given frame type code and in 
 *  a specified directory.
 *
 *  The arguments are:
 *  channel_merge --index-file <ifile> --list-directory <ldir> \
 *                [--rebuild] [--unique-rate] --outfile <ofile> \
 *                <frame-type>
 */
#include <algorithm>
#include <string>
#include <vector>
#include <stdexcept>
#include <stdint.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "chan_list.hh"
#include "lcl_array.hh"
#include "ParseLine.hh"
#include "scandir.hh"
#include "checksum_crc32.hh"
/** The %ChanMerge class is the driver class for the ChanMerge program.
 */
class ChanMerge {
public:
    typedef unsigned long gps_type;
public:
    ChanMerge(int argc, const char* argv[]);
    ~ChanMerge(void);
    void coalesce(void);
    void doit(void);
    void read(void);
    void rebuild(void);
    void sort(void);
    void write_index(const std::string& file);
    void write_index(std::ostream& out);
public:
    class indexEntry {
    public:
	indexEntry(const std::string& name);
	indexEntry(const std::string& dir, const std::string& type, 
		   const sends::epoch& epic);
	~indexEntry(void);
	bool bustup(std::string& dir, std::string& type, std::string& sufx, 
		    sends::epoch& epic) const;
	uint32_t crc32(void) const;
	void calc_crc(void);
	gps_type end_gps(void) const;
	const std::string& file(void) const;
	uint32_t lines(void) const;
	void merge(const indexEntry& ent);
	bool operator<(const indexEntry& ie) const;

	/**  Test for overlapping or adjacent entries.
	 */
	bool overlap(const indexEntry& ent) const;
	const sends::epoch& refEpoch(void) const;
	void set_crc32(uint32_t crc);
	void set_lines(uint32_t nline);
	void set_span(gps_type start, gps_type dt);
	gps_type start_gps(void) const;
    private:
	std::string _fileName;
	sends::epoch _epic;
	uint32_t    _crc32;
	uint32_t    _nlines;
    };

    typedef std::vector<indexEntry>    index_vect;
    typedef index_vect::iterator       index_iter;
    typedef index_vect::const_iterator const_index_iter;

private:
    std::string _frameType;
    std::string _indexFile;
    std::string _listDirectory;
    std::string _outputFile;
    bool        _rebuild;
    bool        _unique_rate;
    index_vect  _indexVector;
};

inline uint32_t
ChanMerge::indexEntry::crc32(void) const {
    return _crc32;
}

inline ChanMerge::gps_type 
ChanMerge::indexEntry::end_gps(void) const {
    return _epic.end();
}

inline const std::string&
ChanMerge::indexEntry::file(void) const {
    return _fileName;
}

inline uint32_t
ChanMerge::indexEntry::lines(void) const {
    return _nlines;
}

inline bool 
ChanMerge::indexEntry::operator<(const indexEntry& ie) const {
    return _epic < ie._epic;
}

inline bool 
ChanMerge::indexEntry::overlap(const indexEntry& ie) const {
    return _crc32 == ie._crc32 && _epic.overlap(ie._epic);
}

inline const sends::epoch&
ChanMerge::indexEntry::refEpoch(void) const {
    return _epic;
}

inline ChanMerge::gps_type 
ChanMerge::indexEntry::start_gps(void) const {
    return _epic.start();
}

using namespace std;
using sends::chan_list;
using sends::chan_entry;
using sends::chan_type;
using sends::data_type;
using sends::epoch;

//======================================  Main progragram entry
int
main(int argc, const char* argv[]) {
    try {
	ChanMerge cm(argc, argv);
	cm.doit();
    } catch (std::exception& e) {
	cerr << "Caught exception: " << e.what() << endl;
    }
}

//======================================  indexEntry Constructor
ChanMerge::indexEntry::indexEntry(const std::string& name) 
    : _fileName(name), _crc32(0), _nlines(0)
{
    string dir, type, sufx;
    bustup(dir, type, sufx, _epic);
}

//======================================  indexEntry Constructor
ChanMerge::indexEntry::indexEntry(const std::string& dir, 
				  const std::string& type, 
				  const sends::epoch& epic) 
{
    ostringstream out;
    out << dir << "/" << type << "-" << epic.start() << "-" 
	<< epic.duration() << ".cls";
    _fileName = out.str();
    _epic   = epic;
    _crc32  = 0;
    _nlines = 0;
}

//======================================  indexEntry Destructor
ChanMerge::indexEntry::~indexEntry(void) {
}

//======================================  Parse channel list file name
bool
ChanMerge::indexEntry::bustup(string& dir, string& type, string& sufx, 
			      sends::epoch& epic) const {
    bool nameok = true;
    string tempstr = _fileName;

    //----------------------------------  Get the directory name
    string::size_type inx = tempstr.rfind('/');
    if (inx != string::npos) {
	dir = tempstr.substr(0, inx);
	tempstr.erase(0, inx+1);
    }
    else {
	dir.clear();
    }

    //-----------------------------------  Get the file extension
    inx = tempstr.rfind('.');
    if (inx != string::npos) {
	sufx = tempstr.substr(inx+1);
	tempstr.erase(inx);
    }
    else {
	sufx.clear();
	nameok = false;
    }

    //-----------------------------------  Pick out the duration.
    sends::gps_type start=0, dt=0;
    inx = tempstr.rfind('-');
    if (inx != string::npos) {
	char* estr = 0;
	dt = strtoul(tempstr.c_str()+inx+1, &estr, 10);
	if (!dt || *estr != 0) {
	    nameok = false;
	}
	tempstr.erase(inx);
    }
    else {
	nameok = false;
    }

    //-----------------------------------  Pick out the start gps
    inx = tempstr.rfind('-');
    if (inx != string::npos) {
	char* estr = 0;
	start = strtoul(tempstr.c_str()+inx+1, &estr, 10);
	if (!dt || *estr != 0) {
	    nameok = false;
	}
	tempstr.erase(inx);
    }
    else {
	nameok = false;
    }
    epic = sends::epoch(start, start+dt);

    //-----------------------------------  The rest is a type code
    type = tempstr;

    return nameok;
}

//======================================  Set index gps times.
void 
ChanMerge::indexEntry::calc_crc(void) {
    //----------------------------------  Open the .cls file
    int fd = ::open(file().c_str(), O_RDONLY);
    if (fd < 0) {
        throw runtime_error(string("calc_crc: Error opening file: ") + file());
    }

    //----------------------------------  Get the file length
    struct stat statBuf;
    ::fstat(fd, &statBuf);
    long lFile = statBuf.st_size;

    //----------------------------------  Read the file
    lcl_array<unsigned char> buf(lFile);
    long length = ::read(fd, buf, lFile);
    ::close(fd);
    if (length != lFile) {
        throw runtime_error("calc_crc: Read length not equal to file length");
    }

    //----------------------------------  Calculate the number of lines
    _nlines = 0;
    for (long i=0; i<lFile; ++i) if (buf[i] == '\n') _nlines++;

    //----------------------------------  Calculate the checksum
    checksum_crc32 crc;
    crc.add(buf, lFile);
    _crc32 = crc.result();
}

//======================================  Set index gps times.
void 
ChanMerge::indexEntry::merge(const indexEntry& ent) {
    if (_epic.overlap(ent._epic) && _epic.duration() < ent._epic.duration()) {
	_fileName = ent._fileName;
    }
    _epic.merge(ent._epic);
}

//======================================  Set index gps times.
void
ChanMerge::indexEntry::set_crc32(uint32_t crc) {
    _crc32 = crc;
}

//======================================  Set index gps times.
void
ChanMerge::indexEntry::set_span(gps_type start, gps_type end) {
    _epic = sends::epoch(start, end);
}

//======================================  Constructor
ChanMerge::ChanMerge(int argc, const char* argv[]) 
    : _rebuild(false), _unique_rate(false)
{
    int j=1;
    bool syntax = false;
    for (int i=1; i<argc && *argv[i] == '-'; i++) {
	string argi = argv[i];
	if (argi == "--index-file") {
	    if (++i < argc) _indexFile = argv[i];
	    else syntax = true;
	}
	else if (argi == "--list-directory") {
	    if (++i < argc) _listDirectory = argv[i];
	    else syntax = true;
	}
	else if (argi == "--rebuild") {
	    _rebuild = true;
	}
	else if (argi == "--outfile") {
	    if (++i < argc) _outputFile = argv[i];
	    else syntax = true;
	}
	else if (argi == "--unique-rate") {
	    _unique_rate = true;
	}
	else {
	    syntax = true;
	    break;
	}
	j = i+1;
    }

    if (!syntax && j<argc) {
	_frameType = argv[j];
    } else {
	cerr << "Error in command line syntax: " << endl;
	cerr << "channel_merge --index-file <ifile> --list-directory <ldir> \\"
	     << endl;
	cerr << "              --rebuild --outfile <ofile> <frame-type>" 
	     << endl;
	throw runtime_error("syntax error in command string");
    }


    if (_indexFile.empty()) {
	ostringstream listdir;
	listdir << _listDirectory << "/" << _frameType 
		<< "-IndexFile.txt";
	_indexFile = listdir.str();
    }
}

//======================================  Destructor
ChanMerge::~ChanMerge(void) {
}

//======================================  Coalesce the index.
void
ChanMerge::coalesce(void) {
    sort();
    size_t N =_indexVector.size();
    size_t j = 0;
    for (size_t i=1; i<N; i++) {
	if (_indexVector[j].overlap(_indexVector[i])) {
	    _indexVector[j].merge(_indexVector[i]);
	} else {
	    j++;
	    if (j < i)  _indexVector[j] = _indexVector[i];
	}
    }
    j++;
    if (j<N) _indexVector.erase(_indexVector.begin()+j, _indexVector.end());
}

//======================================  force a unique rate
static bool
unique_rate(chan_entry& a, chan_entry& b) {
    if (a.name() == b.name() && a.type() == b.type() && a.rate() != b.rate()) {
	size_t na = a.getNSource();
	if (!na) a.clear();
	size_t nb = b.getNSource();
	if (!nb) b.clear();
	if (na && nb) {
	    if (a.source(na-1).end_gps() < b.source(nb-1).end_gps()) {
		cerr << "Channel " << a.name() << " " << a.rate() 
		     << "Hz isonym suppressed! " << endl;
		a.clear();
	    }
	    else {
		cerr << "Channel " << b.name() << " " << b.rate() 
		     << "Hz isonym suppressed! " << endl;
		b.clear();
	    }
	}
	return true;
    }
    return false;
}

//======================================  Processing method
void
ChanMerge::doit(void) {
    if (_rebuild) {
	rebuild();
	coalesce();
	write_index(_indexFile);
    }
    else {
	read();
	coalesce();
    }
    if (_indexVector.empty()) {
	cerr << "Empty index list: " << _indexFile << endl;
	return;
    }

    //----------------------------------  Test active frame type
    string active_file = _listDirectory + "/.active-type";
    bool type_active = (access(active_file.c_str(), F_OK) == 0);
    if (type_active) {
	cout << "# Active frame type: " << _frameType << endl;
    }

    //----------------------------------  Count lines in all index entries.
    //
    //   Count the total number of lines read and the maximum number of lines
    //   in a single entry. Assume that the final index will have no more than 
    //   the maximum 
    size_t N=_indexVector.size();
    size_t total_lines = 0;
    size_t max_lines = 0;
    for (size_t i=0; i<N; ++i) {
	size_t lines_i = _indexVector[i].lines();
	total_lines += lines_i;
	if (lines_i > max_lines) max_lines = lines_i;
    }

    //-----------------------------------  Reserve space for a'best guess' 
    //                                     number of lines.
    size_t reserve_size = 1<<20; // maximum reserve is 1M
    if (reserve_size > total_lines) {
	reserve_size = total_lines;
    }
    else if (reserve_size > 5*max_lines) {
	reserve_size = 5 * max_lines;
    }

    chan_list mergelist;
    mergelist.reserve(reserve_size);

    //----------------------------------  Read in the first file in the list
    epoch epik = _indexVector[0].refEpoch();
    if (type_active && N == 1) epik = epoch(epik.start()); // force infinite end
    mergelist.read(_indexVector[0].file(), epik);

    //----------------------------------  Loop over entries in the index and 
    //                                    read in the file.
    chan_entry null_ce("", chan_type::kUnknown, 0, data_type::kByte1, 0);
    for (size_t i=1; i<N; i++) {

	//------------------------------  Verify there's enough space for file
	size_t lines_i = _indexVector[i].lines();
	if (mergelist.size() + lines_i > mergelist.capacity()) {
	    mergelist.compress(false);
	    size_t need_size = mergelist.size() + lines_i;
	    if (need_size > mergelist.capacity()) mergelist.reserve(need_size);
	}

	//------------------------------  Last epoch of active type doesn't end.
	epik = _indexVector[i].refEpoch();
	if (type_active && i+1 == N) epik = epoch(epik.start());

	//------------------------------  Read channel list.
	mergelist.read(_indexVector[i].file(), epik);
    }
    mergelist.compress();

    //----------------------------------  Force a unique rate:
    //
    //   Loop over all pairs of channels with equal names (indices). 
    //   compare the validity times and remove the earliest isonyms.
    if (_unique_rate) {
	mergelist.iterate_isonym(unique_rate, true);
    }

    //----------------------------------  Sort list
    mergelist.sort();

    //----------------------------------  Dump the file to mergelist.
    if (_outputFile.empty()) {
	mergelist.dump(cout, true);
    } else {
	ofstream outf(_outputFile.c_str());
	mergelist.dump(outf, true);
	outf.close();
    }         
}

//======================================  Read the index file.
void
ChanMerge::read(void) {
    ParseLine pl(_indexFile.c_str());
    if (!pl.isOpen()) {
	string msg = "Unable to open index file: ";
	msg += _indexFile;
	throw runtime_error(msg);
    }

    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
	if (!nArg) continue;
	if (nArg < 4) throw runtime_error("Insufficient entries on line");
	indexEntry ie(pl[0]);
	string   dir, type, sufx;
	sends::epoch epic;
	ie.bustup(dir, type, sufx, epic);
	if (type != _frameType || sufx != "cls") {
	    throw runtime_error(string("Invalid file name: ") + pl[0]);
	}
	ie.set_span(pl.getInt(1), pl.getInt(2));
	ie.set_crc32(pl.getInt(3));
	_indexVector.push_back(ie);
    }
}

//======================================  (re)Build an index
void
ChanMerge::rebuild(void) {
    _indexVector.clear();
    _indexVector.reserve(512);
    
    sends::scandir sd(_listDirectory);
    while (sd.next_entry()) {
	//cout << "file path: " << sd.file_path() << endl;
	if (!sd.is_file()) continue;
	
	indexEntry ie(sd.file_path());
	//cout << "entry: " << ie.file() << "  " << ie.start_gps() << "  " 
	//     << ie.end_gps() << endl;
	string dir, type, sufx;
	sends::epoch epic;
	if (!ie.bustup(dir, type, sufx, epic)) continue;
	//cout << "bustup ok" << endl;
	ie.calc_crc();
	_indexVector.push_back(ie);
    }
}

//======================================  Sort the index
void
ChanMerge::sort(void) {
    if (_indexVector.empty()) return;
    std::sort(_indexVector.begin(), _indexVector.end());
}

//======================================  Write the index to a specified file
void 
ChanMerge::write_index(const std::string& file) {
    ofstream of(file.c_str());
    write_index(of);
}

//======================================  Write the index to a standard ostream
void 
ChanMerge::write_index(std::ostream& out) {
    out << "# Channel list index built by channel merge for frame type: "
	<< _frameType << endl;
    size_t N = _indexVector.size();
    for (size_t i=0; i < N; i++) {
	out <<  _indexVector[i].file() << " " <<  _indexVector[i].start_gps() 
	    << " " <<  _indexVector[i].end_gps() << " " 
	    <<  _indexVector[i].crc32() << endl;
    }
}
